/*
 *   khexedit - Versatile hex editor
 *   Copyright (C) 1999  Espen Sand, espensa@online.no
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <iostream.h>

#include <qdir.h> 
#include <qmsgbox.h> 

#include <kapp.h>
#include <kfiledialog.h>
#include <kurl.h>

#include "hexeditorwidget.h"
#include "hexerror.h"



CHexEditorWidget::CHexEditorWidget( QWidget *parent, const char *name )
  : QWidget( parent, name )
{
  CHexBuffer *hexBuffer = new CHexBuffer;
  if( hexBuffer == 0 ) { return; }
  
  mDocumentList.setAutoDelete( true );
  mDocumentList.append( hexBuffer );
  mUntitledCount = 1;

  mHexView = new CHexViewWidget( this, hexBuffer );
  if( mHexView == 0 || mHexView->widgetValid() == false ) { return; }

  connect( mHexView, SIGNAL(openNewFile()), SLOT( newFile()) );
  connect( kapp, SIGNAL( kdisplayFontChanged() ), SLOT( fontChanged() ) );
  connect( kapp, SIGNAL( kdisplayPaletteChanged() ),SLOT( paletteChanged()) );
  connect( mHexView, SIGNAL( layoutChanged( const SDisplayLayout & ) ),
	   SLOT( layoutChanged( const SDisplayLayout & ) ) );
  connect( mHexView, SIGNAL( inputModeChanged( const SDisplayInputMode & ) ),
	   this, SLOT( inputModeChanged( const SDisplayInputMode & ) ) );

  mHexView->setFocus();
  setBackgroundColor( kapp->windowColor );
   
  mProgressBusy  = false;

  mKfm = 0;
  mGotoDialog    = 0;
  mFindDialog    = 0;
  mFindNavigatorDialog = 0;
  mReplaceDialog = 0;
  mReplacePromptDialog = 0;
  mInsertDialog  = 0;
  mExportDialog  = 0;
  mOptionDialog  = 0;
  mPrintDialog   = 0;
}

CHexEditorWidget::~CHexEditorWidget( void )
{
  delete mHexView;
  delete mGotoDialog;
  delete mFindDialog;
  delete mFindNavigatorDialog;
  delete mReplaceDialog;
  delete mReplacePromptDialog;
  delete mInsertDialog;
  delete mExportDialog;
  delete mOptionDialog;
  delete mPrintDialog;
}


void CHexEditorWidget::initialize( void )
{
  setColor( mDisplayState.color );
  setFont( mDisplayState.font ); 
  mHexView->setMisc( mDisplayState.misc );
  mHexView->setCursor( mDisplayState.cursor, false );
  mHexView->setLayout( mDisplayState.layout );
  mHexView->setInputMode( mDisplayState.input );
  mHexView->setInsertMode( mDisplayState.misc.insertMode );
}


void CHexEditorWidget::writeConfiguration( KConfig &config )
{
  SDisplayLayout &layout = mDisplayState.layout;
  config.setGroup( "Display Options" );
  config.writeEntry( "PrimaryMode",          layout.primaryModeString() );
  config.writeEntry( "SecondaryMode",        layout.secondaryModeString() );
  config.writeEntry( "OffsetMode",           layout.offsetModeString() );
  config.writeEntry( "OffsetVisible",        layout.offsetVisible );
  config.writeEntry( "PrimaryUpperCase",     layout.primaryUpperCase );
  config.writeEntry( "OffsetUpperCase",      layout.offsetUpperCase );
  config.writeEntry( "LineSize",             layout.lineSize );
  config.writeEntry( "ColumnSize",           layout.columnSize );
  config.writeEntry( "LockLine",             layout.lockLine );
  config.writeEntry( "LockColumn",           layout.lockColumn );
  config.writeEntry( "ColumnCharSpace",      layout.columnCharSpace );
  config.writeEntry( "ColumnSpacing",        layout.columnSpacing );
  config.writeEntry( "SeparatorMarginWidth", layout.separatorMarginWidth );
  config.writeEntry( "EdgeMarginWidth",      layout.edgeMarginWidth );
  config.writeEntry( "LeftSeparatorWidth",   layout.leftSeparatorWidth );
  config.writeEntry( "RightSeparatorWidth",  layout.rightSeparatorWidth );


  SDisplayLine &line = mDisplayState.line;
  config.setGroup( "Line Size" );
  config.writeEntry( "Hexadecimal",          line.lineSize[0] );
  config.writeEntry( "Octal",                line.lineSize[1] );
  config.writeEntry( "Binary",               line.lineSize[2] );
  config.writeEntry( "TextOnly",             line.lineSize[3] );
  config.setGroup( "Column Size" );
  config.writeEntry( "Hexadecimal",          line.columnSize[0] );
  config.writeEntry( "Octal",                line.columnSize[1] );
  config.writeEntry( "Binary",               line.columnSize[2] );
  config.writeEntry( "TextOnly",             line.columnSize[3] );

  SDisplayColor &color = mDisplayState.color;
  config.setGroup( "Display Colors" );
  config.writeEntry( "UseSystemColor",       color.useSystemColor );
  config.writeEntry( "TextBg",               color.textBg );
  config.writeEntry( "InactiveBg",           color.inactiveBg );
  config.writeEntry( "PrimaryFg1",           color.primaryFg[0] );
  config.writeEntry( "PrimaryFg2",           color.primaryFg[1] );
  config.writeEntry( "OffsetFg",             color.offsetFg );
  config.writeEntry( "SecondaryFg",          color.secondaryFg );
  config.writeEntry( "MarkBg",               color.markBg );
  config.writeEntry( "MarkFg",               color.markFg );
  config.writeEntry( "LeftSeparatorFg",      color.leftSeparatorFg );
  config.writeEntry( "RightSeparatorFg",     color.leftSeparatorFg );
  config.writeEntry( "NonPrintFg",           color.nonPrintFg );
  config.writeEntry( "CursorBg",             color.cursorBg );
  config.writeEntry( "CursorFg",             color.cursorFg );

  SDisplayFont &font = mDisplayState.font;
  config.setGroup( "Display Font" );
  config.writeEntry( "UseSystemFont",        font.useSystemFont );
  config.writeEntry( "LocalFont",            font.localFont );
  config.writeEntry( "NonPrintChar",         font.nonPrintChar );

  SDisplayCursor &cursor = mDisplayState.cursor;
  config.setGroup( "Display Cursor" );
  config.writeEntry( "NeverBlink",           cursor.alwaysVisible );
  config.writeEntry( "AlwaysBlockShape",     cursor.alwaysBlockShape );
  config.writeEntry( "ThickInsertShape",     cursor.thickInsertShape );
  config.writeEntry( "FocusMode",            cursor.focusModeString() );
  config.writeEntry( "Interval",             cursor.interval );

  SDisplayMisc &misc = mDisplayState.misc;
  config.setGroup( "Display Misc" );
  config.writeEntry( "UndoLevel",            misc.undoLevel );
  config.writeEntry( "OpenFile",             misc.openFileString() );
  config.writeEntry( "InputSound",           misc.inputSound );
  config.writeEntry( "FatalSound",           misc.fatalSound );
  config.writeEntry( "AutoCopyToClipboard",  misc.autoCopyToClipboard );
  config.writeEntry( "InsertMode",           misc.insertMode );
  config.writeEntry( "WriteProtect",         misc.writeProtect );
  config.writeEntry( "ConfirmWrap",          misc.confirmWrap );
  config.writeEntry( "CursorJump",           misc.cursorJump );
  config.writeEntry( "MakeBackup",           misc.makeBackup );
  config.writeEntry( "ConfirmThreshold",     misc.confirmThreshold );
  config.writeEntry( "ThresholdValue",       misc.thresholdValue );

  if( mPrintDialog != 0 )
  {
    mPrintDialog->writeConfiguration();
  }

}

void CHexEditorWidget::readConfiguration( KConfig &config )
{
  SDisplayLayout &layout = mDisplayState.layout;
  config.setGroup( "Display Options" );
  layout.setPrimaryMode( config.readEntry("PrimaryMode") );
  layout.setSecondaryMode( config.readEntry("SecondaryMode") );
  layout.setOffsetMode( config.readEntry("OffsetMode") );
  layout.offsetVisible = config.readBoolEntry( 
    "OffsetVisible", layout.offsetVisible );
  layout.primaryUpperCase = config.readBoolEntry( 
    "PrimaryUpperCase", layout.primaryUpperCase );
  layout.offsetUpperCase = config.readBoolEntry( 
    "OffsetUpperCase", layout.offsetUpperCase );
  layout.lineSize = config.readNumEntry( 
    "LineSize", layout.lineSize );
  layout.columnSize = config.readNumEntry(
    "ColumnSize", layout.columnSize );
  layout.lockLine = config.readBoolEntry(
    "LockLine", layout.lockLine );
  layout.lockColumn = config.readBoolEntry( 
    "LockColumn", layout.lockColumn );
  layout.columnCharSpace = config.readBoolEntry( 
    "ColumnCharSpace", layout.columnCharSpace );
  layout.columnSpacing = config.readNumEntry( 
    "ColumnSpacing", layout.columnSpacing );
  layout.separatorMarginWidth = config.readNumEntry( 
    "SeparatorMarginWidth", layout.separatorMarginWidth );
  layout.edgeMarginWidth = config.readNumEntry( 
    "EdgeMarginWidth", layout.edgeMarginWidth );
  layout.leftSeparatorWidth = config.readNumEntry( 
    "LeftSeparatorWidth", layout.leftSeparatorWidth );
  layout.rightSeparatorWidth = config.readNumEntry( 
    "RightSeparatorWidth", layout.rightSeparatorWidth );

  SDisplayLine &line = mDisplayState.line;
  config.setGroup( "Line Size" );
  line.lineSize[0] = config.readUnsignedNumEntry(
    "Hexadecimal", line.lineSize[0] );
  line.lineSize[1] = config.readUnsignedNumEntry(
    "Octal", line.lineSize[1] );
  line.lineSize[2] = config.readUnsignedNumEntry(
    "Binary", line.lineSize[2] );
  line.lineSize[3] = config.readUnsignedNumEntry(
    "TextOnly", line.lineSize[3] );
  config.setGroup( "Column Size" );
  line.columnSize[0] = config.readUnsignedNumEntry(
    "Hexadecimal", line.columnSize[0] );
  line.columnSize[1] = config.readUnsignedNumEntry(
    "Octal", line.columnSize[1] );
  line.columnSize[2] = config.readUnsignedNumEntry(
    "Binary", line.columnSize[2] );
  line.columnSize[3] = config.readUnsignedNumEntry(
    "TextOnly", line.columnSize[3] );

  SDisplayFont &font = mDisplayState.font;
  config.setGroup( "Display Font" );
  font.useSystemFont = config.readBoolEntry( 
    "UseSystemFont", font.useSystemFont );
  font.localFont = config.readFontEntry( 
    "LocalFont", &font.localFont );
  font.nonPrintChar = config.readNumEntry( 
    "NonPrintChar", font.nonPrintChar );

  SDisplayColor &color = mDisplayState.color;
  config.setGroup( "Display Colors" );
  color.useSystemColor = config.readBoolEntry( 
    "UseSystemColor", color.useSystemColor );
  color.textBg = config.readColorEntry( 
    "TextBg", &color.textBg );
  color.inactiveBg = config.readColorEntry( 
    "InactiveBg", &color.inactiveBg );
  color.primaryFg[0] = config.readColorEntry(
    "PrimaryFg1", &color.primaryFg[0] );
  color.primaryFg[1] = config.readColorEntry( 
    "PrimaryFg2", &color.primaryFg[1] );
  color.offsetFg = config.readColorEntry( 
    "OffsetFg", &color.offsetFg );
  color.secondaryFg = config.readColorEntry(
    "SecondaryFg", &color.secondaryFg );
  color.markBg = config.readColorEntry(
    "MarkBg", &color.markBg );
  color.markFg = config.readColorEntry( 
    "MarkFg", &color.markFg );
  color.cursorBg = config.readColorEntry( 
    "CursorBg", &color.cursorBg );
  color.cursorFg = config.readColorEntry( 
    "CursorFg", &color.cursorFg );
  color.leftSeparatorFg = config.readColorEntry( 
    "LeftSeparatorFg", &color.leftSeparatorFg );
  color.rightSeparatorFg = config.readColorEntry( 
    "RightSeparatorFg", &color.rightSeparatorFg );
  color.nonPrintFg = config.readColorEntry( 
    "NonPrintFg", &color.nonPrintFg );


  SDisplayCursor &cursor = mDisplayState.cursor;
  config.setGroup( "Display Cursor" );
  cursor.alwaysVisible = config.readBoolEntry( 
    "NeverBlink", cursor.alwaysVisible );
  cursor.alwaysBlockShape = config.readBoolEntry( 
    "AlwaysBlockShape", cursor.alwaysBlockShape );
  cursor.thickInsertShape = config.readBoolEntry( 
    "ThickInsertShape", cursor.thickInsertShape );
  cursor.setFocusMode( config.readEntry("FocusMode") );
  cursor.interval = config.readNumEntry( 
    "Interval", cursor.interval );

  SDisplayMisc &misc = mDisplayState.misc;
  config.setGroup( "Display Misc" );
  misc.undoLevel = config.readNumEntry( 
    "UndoLevel", misc.undoLevel );
  misc.setOpenFile( config.readEntry("OpenFile") );
  misc.inputSound = config.readBoolEntry( 
    "InputSound", misc.inputSound );
  misc.fatalSound = config.readBoolEntry( 
    "FatalSound", misc.fatalSound );
  misc.autoCopyToClipboard = config.readBoolEntry( 
    "AutoCopyToClipboard", misc.autoCopyToClipboard );
  misc.insertMode = config.readBoolEntry( 
    "InsertMode", misc.insertMode );
  misc.writeProtect = config.readBoolEntry( 
    "WriteProtect", misc.writeProtect );
  misc.confirmWrap = config.readBoolEntry( 
    "ConfirmWrap", misc.confirmWrap );
  misc.cursorJump = config.readBoolEntry( 
    "CursorJump", misc.cursorJump );
  misc.makeBackup = config.readBoolEntry( 
    "MakeBackup", misc.makeBackup );
  misc.confirmThreshold = config.readBoolEntry( 
    "ConfirmThreshold", misc.confirmThreshold );
  misc.thresholdValue = config.readNumEntry( 
    "ThresholdValue", misc.thresholdValue );
}



void CHexEditorWidget::resizeEvent( QResizeEvent *e )
{
  mHexView->resize( e->size().width(), e->size().height() );
}


void CHexEditorWidget::fontChanged( void )
{
  if( mDisplayState.font.useSystemFont == true )
  {
    setFont( mDisplayState.font );
  }
}

void CHexEditorWidget::paletteChanged( void )
{
  setColor( mDisplayState.color );
}

void CHexEditorWidget::layoutChanged( const SDisplayLayout &layout )
{
  //mDisplayState.layout = layout;
}

void CHexEditorWidget::inputModeChanged( const SDisplayInputMode &input )
{
  mDisplayState.input = input;
}


void CHexEditorWidget::setLineSize(const SDisplayLine &line )
{
  mDisplayState.line = line;
}

void CHexEditorWidget::setLayout( const SDisplayLayout &layout )
{
  //
  // We only set the values that can be modified by the dialog
  //
  mDisplayState.layout.lockLine = layout.lockLine;
  mDisplayState.layout.lockColumn = layout.lockColumn;
  mDisplayState.layout.leftSeparatorWidth = layout.leftSeparatorWidth;
  mDisplayState.layout.rightSeparatorWidth = layout.rightSeparatorWidth;
  mDisplayState.layout.separatorMarginWidth = layout.separatorMarginWidth;
  mDisplayState.layout.edgeMarginWidth = layout.edgeMarginWidth;
  mDisplayState.layout.columnCharSpace = layout.columnCharSpace;
  mDisplayState.layout.columnSpacing = layout.columnSpacing;
  
  //
  // Select the line and column sizes we shall use now.
  //
  SDisplayLayout &l = mDisplayState.layout;
  l.lineSize   = mDisplayState.line.lineSize[ l.primaryMode ];
  l.columnSize = mDisplayState.line.columnSize[ l.primaryMode ];

  mHexView->setLayout( mDisplayState.layout );
}
  
void CHexEditorWidget::setCursor( const SDisplayCursor &cursor )
{
  mDisplayState.cursor = cursor;
  mHexView->setCursor( mDisplayState.cursor, true );
}

void CHexEditorWidget::setColor( const SDisplayColor &color )
{
  mDisplayState.color = color;

  mDisplayState.color.selectBg = kapp->selectColor;
  mDisplayState.color.selectFg = kapp->selectTextColor;

  if( mDisplayState.color.useSystemColor == true )
  {
    SDisplayColor c = mDisplayState.color;
    c.textBg = kapp->windowColor;
    c.inactiveBg = kapp->windowColor;
    c.primaryFg[0] = kapp->windowTextColor;
    c.primaryFg[1] = kapp->windowTextColor;
    c.nonPrintFg = kapp->windowTextColor;
    c.offsetFg = kapp->windowTextColor;
    c.secondaryFg = kapp->windowTextColor;
    c.leftSeparatorFg = kapp->windowTextColor;
    c.rightSeparatorFg = kapp->windowTextColor;
    c.cursorBg = kapp->windowTextColor;
    c.cursorFg = kapp->windowColor;
    mHexView->setColor( c, true ); 
  }
  else
  {
    mHexView->setColor( mDisplayState.color, true ); 
  }
}


void CHexEditorWidget::setFont( const SDisplayFont &font )
{
  mDisplayState.font = font;

  SDisplayFontInfo fontInfo;
  if( mDisplayState.font.useSystemFont == true )
  {
    fontInfo.font = kapp->fixedFont;
  }
  else
  {
    fontInfo.font = mDisplayState.font.localFont;
  }
  fontInfo.nonPrintChar = mDisplayState.font.nonPrintChar;
  mHexView->setFont( fontInfo, true );
}

void CHexEditorWidget::setMisc( const SDisplayMisc &misc )
{
  mDisplayState.misc = misc;
  mHexView->setMisc( mDisplayState.misc );
}



void CHexEditorWidget::newFile( void )
{
  const char *newName = i18n( "Untitled" );
  QString url( strlen( newName ) + 20 );
  url.sprintf( "%s %u", newName, mUntitledCount );
  
  //
  // If the url is already present in the document list (should not happen), 
  // then this document is "raised". 
  //
  bool success = selectDocument( url );
  if( success == true )
  {
    return;
  }
  
  //
  // Prepare a new buffer we can load the document into
  //
  success = createBuffer();
  if( success == false )
  {
    return;
  }

  int errCode = mHexView->newFile( url );
  if( errCode != Err_Success )
  {
    QMessageBox::warning(
      topLevelWidget(),
      i18n("Operation failed"),
      i18n("Unable to create new document."),
      i18n("&OK"),
      "",
      "",
      0,0 );
    return;
  }

  mUntitledCount += 1;
}


void CHexEditorWidget::open( void )
{
  QString url = KFileDialog::getOpenFileURL( 0, "*", topLevelWidget() );
  if( url.isEmpty() == true )
  {
    return;
  }

  //
  // If the url is already present in the document list, then this
  // document is "raised". 
  //
  bool success = selectDocument( url );
  if( success == true )
  {
    return;
  }

  //
  // Prepare a new buffer we can load the document into
  //
  success = createBuffer();
  if( success == false )
  {
    return;
  }

  int errCode = readURL( url, false );
  if( errCode != Err_Success )
  {
    emit errorLoadFile( url );
    removeBuffer();
  }

  defaultWriteProtection();
}



void CHexEditorWidget::open( const QString &url )
{
  //
  // If the url is already present in the document list, then this
  // document is "raised". 
  //
  bool success = selectDocument( url );
  if( success == true )
  {
    return;
  }
  
  //
  // Prepare a new buffer we can load the document into
  //
  success = createBuffer();
  if( success == false )
  {
    return;
  }

  int errCode = readURL( url, false );
  if( errCode != Err_Success )
  {
    emit errorLoadFile( url );
    removeBuffer();
  }

  defaultWriteProtection();
}




bool CHexEditorWidget::selectDocument( const QString &url )
{
  CHexBuffer *hexBuffer = documentItem( url );
  if( hexBuffer == 0 )
  {
    return( false );
  }

  mHexView->setBuffer( hexBuffer );
  return( true );
}


void CHexEditorWidget::insertFile( void )
{
  QString url = KFileDialog::getOpenFileURL( 0, "*", topLevelWidget(), 0 );
  if( url.isEmpty() == true )
  {
    return;
  }

  readURL( url, true );
}



void CHexEditorWidget::stop( void )
{
  mProgressStop = mProgressBusy;
}


bool CHexEditorWidget::close( void )
{
  if( querySave() == false )
  {
    return( false );
  }

  removeBuffer();
  return( true );
}




bool CHexEditorWidget::closeAll( void )
{
  while( mDocumentList.count() > 0 )
  {
    CHexBuffer *buf;
    for( buf = mDocumentList.first(); buf != 0; buf = mDocumentList.next() )
    {
      //
      // We select the buffer we are about to close becuase it will
      // the become visible, and because 'close()' works on the current
      // buffer.
      //
      mHexView->setBuffer( buf );
      if( close() == false )
      {
	return( false );
      }
      break;
    }

    //
    // The document list will always contain at least one element because
    // the hexview widget needs a hexbuffer to work. If this hexbuffer is
    // not valid, then all files have been closed.
    //
    if( mDocumentList.count() == 1 && mHexView->valid() == false )
    {
      break;
    }
  }

  return( true );
}



bool CHexEditorWidget::querySave( void )
{
  if( mHexView->modified() == false )
  {
    return( true );
  }

  QString msg;
  msg += i18n( "The current document has been modified." );
  msg += "\n";
  msg += i18n( "Do you want to save it?" );
      
  int reply = QMessageBox::warning( 
    topLevelWidget(),
    i18n( "Warning" ), 
    msg.data(),
    i18n( "&Yes" ),
    i18n( "&No" ),
    i18n( "&Cancel" ),
    0,2 );
  if( reply == 0 ) // Yes
  {
    save();
  }
  else if( reply == 1 )
  {
  }
  else
  {
    return( false );
  }

  return( true );
}


bool CHexEditorWidget::backup( void )
{
  if( mHexView->valid() == false || mHexView->urlValid() == false )
  {
    return( false );
  }

  KURL kurl( mHexView->url() );
  if( kurl.isLocalFile() == false )
  {
    return( false );
  }
    
  QString curName( kurl.path() );
  KURL::decodeURL( curName );

  QString newName( curName );
  newName += "~";

  int errCode = rename( curName, newName );
  if( errCode != 0 )
  {
    return( false );
  }

  return( true );
}






void CHexEditorWidget::save( void )
{
  if( mHexView->valid() == false )
  {
    return;
  }
  if( mHexView->urlValid() == true )
  {
    if( mDisplayState.misc.makeBackup == true )
    {
      backup();
    }
    writeURL( mHexView->url() ); 
  }
  else
  {
    saveAs();
  }
}

void CHexEditorWidget::saveAs( void )
{ 
  if( mHexView->valid() == false )
  {
    return;
  }

  QString url; 
  int query = 0;
  do 
  {
    url = KFileDialog::getSaveFileURL( mHexView->url().data(), "*", this );
    if( url.isEmpty() == true )
    {
      return;
    }

    KURL u(url);
    if( u.isLocalFile() == true ) 
    {
      QString name( u.path() );
      KURL::decodeURL( name );
      QFileInfo info( name );

      if( info.exists() == true ) 
      {
	QString msg;
	msg += i18n( "A document with this name already exists." );
	msg += "\n";
	msg += i18n( "Do you want to overwrite it?" );

        query = QMessageBox::warning(
          this,
          i18n( "Warning" ),
	  msg.data(),
          i18n( "&Yes" ),
          i18n( "&No" ),
          "",
	  0, 1 );
      }
    }
  } 
  while( query == 1 );

  mHexView->setUrl( url );
  writeURL( mHexView->url() ); 

}



void CHexEditorWidget::reload( void )
{
  if( mHexView->valid() == false )
  {
    return;
  }

  if( mHexView->urlValid() == false )
  {
    QString msg;
    msg += i18n( "The current document does not exist on disk." );
  
    QMessageBox::warning(
      topLevelWidget(),
      i18n( "Reload" ),
      msg.data(),
      i18n( "&OK" ),
      "",
      "",
      0, 0 );
    return;
  }

  if( mHexView->modified() == true )
  {
    QString msg;
    msg += i18n( "The current document has been modified." );
    msg += "\n";
    msg += i18n( "Do you still want to reload it?" );

    int reply = QMessageBox::warning( 
      topLevelWidget(),
      i18n( "Warning" ), 
      msg.data(),
      i18n( "&Yes" ),
      i18n( "&No" ),
      0, 0 );
    if( reply == 1 ) // No
    {
      return;
    }
  }

  QString url( mHexView->url().data() );
  mHexView->closeFile();
  readURL( url, false );
}

void CHexEditorWidget::print( void )
{
  if( mPrintDialog == 0 )
  {
    const char *title = i18n( "Print active document" );
    mPrintDialog = new CPrinterDialog( title, 0 );
    if( mPrintDialog == 0 ) { return; }
    connect( mPrintDialog, SIGNAL( printPostscript(CHexPrinter &)),
	     SLOT( printPostscript( CHexPrinter &)) );
    connect( mPrintDialog, SIGNAL( printText(CHexPrinter &, const QString &)),
	     SLOT( printText( CHexPrinter &, const QString &)) );
  }
  if( mPrintDialog->isVisible() == false )
  {
    centerDialog( mPrintDialog, topLevelWidget() );
    mPrintDialog->show();
  }
  else
  {
    mPrintDialog->raise();
  }
}




void CHexEditorWidget::printPostscript( CHexPrinter &printer )
{
  if( mDisplayState.misc.confirmThreshold == true )
  {
    uint numPage = mHexView->numPage( printer );
    if( numPage > mDisplayState.misc.thresholdValue )
    {
      QString msg;
      msg.sprintf( "%s\n\n%s %d %s\n%s",
		   i18n("Print threshold excceded."), 
		   i18n("You are about to print"),  numPage, i18n("pages."),
		   i18n("Continue?") );

      int query = QMessageBox::warning(
        this,
	i18n( "Print" ),
	msg.data(),
	i18n( "&Yes" ),
	i18n( "&No" ),
	"",
	0, 1 );
      if( query == 1 )
      {
	return;
      }
    }
  }

  //
  // No i18n() on this one! 
  // Constants come from config.h
  //
  QString creator( PACKAGE );
  creator += " ";
  creator += VERSION;

  printer.setCreator( creator );

  int errCode = prepareProgressData( pg_print );
  if( errCode == Err_Success )
  {
    errCode = mHexView->print( printer, mProgressData );
  }
 
  QString msg( i18n("Could not print data."));
  msg += "\n";
  msg += hexError( errCode );

  if( errCode != Err_Success )
  {
    QMessageBox::warning(
      topLevelWidget(),
      i18n("Print"),
      msg.data(),
      i18n("&OK"),
      "",
      "",
      0,0 );
  }
}






void CHexEditorWidget::printText( CHexPrinter &printer, const QString &mode )
{
  SExportRange range;
  if( printer.all() == true )
  {
    range.mode = SExportRange::All;
  }
  else if( printer.selection() == true )
  {
    range.mode = SExportRange::Selection;
  }
  else
  {
    range.mode  = SExportRange::Range;
    range.start = printer.startRange();
    range.stop  = printer.stopRange();
  }

  QString header;
  if( mode.isEmpty() == false )
  {
    header = mode;
  }
  else
  {
    header = i18n("Print");
  }

  if( printer.outputToFile() == false )
  {
    QMessageBox::warning(
      topLevelWidget(),
      header.data(),
      i18n("Can not yet print plain text to printer"),
      i18n("&OK"),
      "",
      "",
      0, 0);
    return;
  }

  QFile file;
  if( printer.outputToStdout() == true )
  {
    if( file.open( IO_WriteOnly, stdout ) == false )
    {
      QString msg = i18n( "Unable to open file for writing" );
      msg += "\n";
      msg += i18n("stdout");

      QMessageBox::warning(
        topLevelWidget(),
        header.data(),
        msg.data(),
        i18n("&OK"),
        "",
        "",
        0,0 );
      return;
    }
  }
  else
  {
    file.setName( printer.outputFileName() );
    if( file.open( IO_WriteOnly ) == false )
    {
      QString msg = i18n( "Unable to open file for writing" );
      msg += "\n";
      msg += file.name();

      QMessageBox::warning(
        topLevelWidget(),
        header.data(),
        msg.data(),
        i18n("&OK"),
        "",
        "",
        0,0 );
      return;
    }
  }

  int errCode = prepareProgressData( pg_print );
  if( errCode == Err_Success )
  {
    errCode = mHexView->export( file, range, mProgressData );
  }
 
  if( errCode != Err_Success )
  {
    QString msg( i18n("Unable to write data."));
    msg += "\n";
    msg += hexError( errCode );

    QMessageBox::warning(
      topLevelWidget(),
      header.data(),
      msg.data(),
      i18n("&OK"),
      "",
      "",
      0,0 );
  }

  file.close();
}




void CHexEditorWidget::export( void )
{
  if( mExportDialog == 0 )
  {
    mExportDialog = new CExportDialog( 0, "Export" );
    if( mExportDialog == 0 ) { return; }
    mExportDialog->setCaption( i18n("Export") );

    connect( mExportDialog, SIGNAL( export(CHexPrinter &, const QString &)),
	     SLOT( printText( CHexPrinter &, const QString &)) );
  }
  if( mExportDialog->isVisible() == false )
  {
    centerDialog( mExportDialog, topLevelWidget() );
    mExportDialog->defaultFocus();
    mExportDialog->show();
  }
  else
  {
    mExportDialog->raise();
  }
}



void CHexEditorWidget::encode( CConversion::EMode mode )
{
  int errCode = prepareProgressData( pg_encode );
  if( errCode != Err_Success )
  {
    QMessageBox::warning(
      topLevelWidget(),
      i18n("Encoding failed"),
      i18n("Something went wrong."),
      i18n("&OK"),
      "",
      "",
      0,0 );
  }

  mHexView->setEncoding( mode, mProgressData );
}





void CHexEditorWidget::undo( void )
{
  mHexView->undo();
}

void CHexEditorWidget::redo( void )
{
  mHexView->redo();
}

void CHexEditorWidget::toggleWriteProtection( void )
{
  mDisplayState.input.readOnly = mDisplayState.input.readOnly == true ? 
    false : true;
  mHexView->setInputMode( mDisplayState.input );
}

void CHexEditorWidget::defaultWriteProtection( void )
{
  mDisplayState.input.readOnly = mDisplayState.misc.writeProtect; 
  mHexView->setInputMode( mDisplayState.input );
}


void CHexEditorWidget::benchmark( void )
{
  mHexView->benchmark();
}

void CHexEditorWidget::copy( void )
{
  mHexView->copy();
}

void CHexEditorWidget::copyPrimaryFormatted( void )
{
  mHexView->copyFormatted( false );
}

void CHexEditorWidget::copyAllFormatted( void )
{
  mHexView->copyFormatted( true );
}

void CHexEditorWidget::paste( void )
{
  if( mHexView->valid() == false )
  {
    pasteNewFile();
  }
  else
  {
    mHexView->paste();
  }
}

void CHexEditorWidget::pasteNewFile( void )
{
  newFile();
  mHexView->paste();
}

void CHexEditorWidget::cut( void )
{
  mHexView->cut();
}

void CHexEditorWidget::selectAll( void )
{
  mHexView->selectAll();
}

void CHexEditorWidget::unselect( void )
{
  mHexView->unselect();
}


void CHexEditorWidget::addBookmark( void )
{
  mHexView->addBookmark( -1 );
}

void CHexEditorWidget::removeBookmark( bool all )
{
  if( mHexView->bookmarkCount() == 0 )
  {
    return;
  }

  if( all == true )
  {
    QString msg;
    msg += i18n( "Deleted bookmarks can not be restored." );
    msg += "\n";
    msg += i18n( "Proceed?" );

    int reply = QMessageBox::information( 
      topLevelWidget(),
      i18n( "Warning" ), 
      msg.data(),
      i18n( "&Yes" ),
      i18n( "&No" ),
      0, 
      0, 1 );
    if( reply == 1 ) // No
    {
      return;
    }
  }

  mHexView->removeBookmark( all );
}      

void CHexEditorWidget::replaceBookmark( void )
{
  mHexView->replaceBookmark();
}      

void CHexEditorWidget::gotoBookmark( uint position )
{
  mHexView->gotoBookmark( position );
}   




void CHexEditorWidget::gotoOffset( void )
{
  if( mGotoDialog == 0 )
  {
    mGotoDialog = new CGotoDialog( 0, "goto" );
    if( mGotoDialog == 0 ) { return; }
    mGotoDialog->setCaption( i18n("Goto") );
    connect( mGotoDialog, SIGNAL(gotoOffset( uint, uint, bool, bool )), 
	     mHexView, SLOT(gotoOffset( uint, uint, bool, bool )) );
  }
  if( mGotoDialog->isVisible() == false )
  {
    centerDialog( mGotoDialog, topLevelWidget() );
    mGotoDialog->defaultFocus();
    mGotoDialog->show();
  }
  else
  {
    mGotoDialog->raise();
  }
}


void CHexEditorWidget::find( void )
{
  if( mFindNavigatorDialog != 0 )
  {
    mFindNavigatorDialog->hide();
  }

  if( mFindDialog == 0 )
  {
    mFindDialog = new CFindDialog( 0, "find" );
    if( mFindDialog == 0 ) { return; }
    mFindDialog->setCaption( i18n("Find") );
    connect( mFindDialog, 
	     SIGNAL(findData(SSearchControl &, uint, bool)),
	     SLOT(findData(SSearchControl &, uint, bool)) );
  }
  if( mFindDialog->isVisible() == false )
  {
    centerDialog( mFindDialog, topLevelWidget() );
    mFindDialog->defaultFocus();
    mFindDialog->show();
  }
  else
  {
    mFindDialog->raise();
  }
}



void CHexEditorWidget::findData( SSearchControl &sc, uint mode, bool navigator)
{
  for( uint i=0; i < 2; i++ )
  {
    int errCode;
    if( mode == Find_First )
    {
      errCode = mHexView->findFirst( sc );
      if( errCode == Err_Success )
      {
	if( navigator == true )
	{
	  findNavigator( sc );
	}
	return;
      }
    }
    else if( mode == Find_Next )
    {
      errCode = mHexView->findNext( sc );
      if( errCode == Err_Success )
      {
	if( navigator == true )
	{
	  findNavigator( sc );
	}
	return;
      }
    }
    else
    {
      return;
    }

    if( errCode == Err_WrapBuffer && i == 0 )
    {
      bool reply = askWrap( sc.forward, QString(i18n("Find")) );
      if( reply == false )
      {
	return;
      }
      
      sc.fromCursor = false;
      mode = Find_First;
    }
  }

  if( mode == Find_First )
  {
    QString msg;
    msg += i18n( "Search key not found in document." );

    QMessageBox::information(
      topLevelWidget(),
      i18n( "Find" ), 
      msg.data(),
      i18n( "&OK" ),
      "",
      "",
      0, 0 );
  }
}



void CHexEditorWidget::findAgain( void )
{
  if( canFind( true ) == true )
  {
    mFindDialog->findAgain( CFindDialog::find_Again );
  }
}


void CHexEditorWidget::findNext( void )
{
  if( canFind( true ) == true )
  {
    mFindDialog->findAgain( CFindDialog::find_Next );
  }
}


void CHexEditorWidget::findPrevious( void )
{
  if( canFind( true ) == true )
  {
    mFindDialog->findAgain( CFindDialog::find_Previous );
  }
}


bool CHexEditorWidget::askWrap( bool fwd, const QString &header )
{
  QString msg;
  if( fwd == true )
  {
    msg += i18n( "End of document reached." );
    msg += "\n";
    msg += i18n( "Continue from the beginning?" );
  }
  else
  {
    msg += i18n( "Beginning of document reached." );
    msg += "\n";
    msg += i18n( "Continue from the end?" );
  }
 
  int reply = QMessageBox::information(
    topLevelWidget(),
    header.data(),
    msg.data(),
    i18n( "&Yes" ),
    i18n( "&No" ),
    "",
    0, 1 );
  return( reply == 1 ? false : true );
}


bool CHexEditorWidget::canFind( bool showError )
{
  if( mFindDialog == 0 || mFindDialog->data() == 0 )
  {
    if( showError == true )
    {
      QString msg;
      msg += i18n("Your request can not be processed.");
      msg += "\n";
      msg += i18n("No search pattern is defined.");
 
      QMessageBox::warning(
        topLevelWidget(),
	i18n("Find"),
	msg.data(),
	i18n("&OK"),
	"",
	"",
	0,0 );
    }
    return( false );
  }

  return( true );
}



void CHexEditorWidget::findNavigator( SSearchControl &sc )
{
  if( canFind( false ) == false )
  {
    return;
  }

  if( mFindNavigatorDialog == 0 )
  {
    mFindNavigatorDialog = new CFindNavigatorDialog( 0, "find" );
    if( mFindNavigatorDialog == 0 ) { return; }
    mFindNavigatorDialog->setCaption( i18n("Find (Navigator)") );
    connect( mFindNavigatorDialog, 
	     SIGNAL(findData(SSearchControl &, uint, bool)),
	     SLOT(findData(SSearchControl &, uint, bool)) );
    connect( mFindNavigatorDialog, SIGNAL(makeKey(void)), 
	     SLOT(find()) );
  }
  if( mFindNavigatorDialog->isVisible() == false )
  {
    centerDialogBottom( mFindNavigatorDialog, topLevelWidget() );
    mFindNavigatorDialog->defineData( sc );
    mFindNavigatorDialog->show();
  }
  else
  {
    mFindNavigatorDialog->raise();
  }
}



void CHexEditorWidget::replace( void )
{
  hideReplacePrompt();

  if( mReplaceDialog == 0 )
  {
    mReplaceDialog = new CReplaceDialog( 0, "replace" );
    if( mReplaceDialog == 0 ) { return; }
    mReplaceDialog->setCaption( i18n("Find and Replace") );
    connect( mReplaceDialog, 
	     SIGNAL( replaceData( SSearchControl &, uint)),
	     SLOT( replaceData( SSearchControl &, uint)));
  }
  if( mReplaceDialog->isVisible() == false )
  {
    centerDialog( mReplaceDialog, topLevelWidget() );
    mReplaceDialog->defaultFocus();
    mReplaceDialog->show();
  }
  else
  {
    mReplaceDialog->raise();
  }
}



void CHexEditorWidget::replaceData( SSearchControl &sc, uint mode )
{
  while( 1 )
  {
    if( mode == Replace_All || mode == Replace_AllInit )
    {
      mHexView->replaceAll( sc, mode == Replace_AllInit ? true : false );
    }
    else if( mode == Replace_First )
    {
      int errCode = mHexView->findFirst( sc );
      if( errCode == Err_Success )
      {
	replacePrompt( sc );
	return;
      }
    }
    else if( mode == Replace_Next )
    {
      int errCode = mHexView->replaceMarked( sc );
      if( errCode != Err_Success )
      {
	// Perhaps a notification here ?
      }
      errCode = mHexView->findNext( sc );
      if( errCode == Err_Success )
      {
	replacePrompt( sc );
	return;
      }
    }
    else if( mode == Replace_Ignore )
    {
      int errCode = mHexView->findNext( sc );
      if( errCode == Err_Success )
      {
	replacePrompt( sc );
	return;
      }
    }
    else
    {
      break;
    }

    if( sc.wrapValid == false )
    {
      break;
    }

    bool reply = askWrap( sc.forward, QString(i18n("Find and Replace")) );
    if( reply == false )
    {
      break;
    }

    int errCode = mHexView->findWrap( sc );
    if( errCode != Err_Success )
    {
      break;
    }
    
    if( mode == Replace_All || mode == Replace_AllInit )
    {
      mode = Replace_All;
      continue;
    }

    replacePrompt( sc );
    return;
  }

  replaceResult( sc );
}

void CHexEditorWidget::replacePrompt( SSearchControl &sc )
{
  if( mReplacePromptDialog == 0 )
  {
    mReplacePromptDialog = new CReplacePromptDialog( 0, "replace" );
    if( mReplacePromptDialog == 0 ) { return; }
    mReplacePromptDialog->setCaption( i18n("Find and Replace") );
    connect( mReplacePromptDialog, 
	     SIGNAL( replaceData( SSearchControl &, uint)),
	     SLOT( replaceData( SSearchControl &, uint)));
  }
  if( mReplacePromptDialog->isVisible() == false )
  {
    centerDialogBottom( mReplacePromptDialog, topLevelWidget() );
    mReplacePromptDialog->defineData( sc );
    mReplacePromptDialog->show();
  }
  else
  {
    mReplacePromptDialog->raise();
  }

  //mHexView->setEnabled( false );
}


void CHexEditorWidget::hideReplacePrompt( void )
{
  if( mReplacePromptDialog != 0 )
  {
    mReplacePromptDialog->hide();
  }

  //mHexView->setEnabled( true );
}



void CHexEditorWidget::replaceResult( SSearchControl &sc )
{
  hideReplacePrompt();

  if( sc.match == false )
  {
    QString msg;
    if( sc.inSelection == true )
    {
      msg += i18n( "Search key not found in selected area." );
    }
    else
    {
      msg += i18n( "Search key not found in document." );
    }

    QMessageBox::information(
      topLevelWidget(),
      i18n( "Find and Replace" ), 
      msg.data(),
      i18n( "&OK" ),
      "",
      "",
      0, 0 );
  }
  else
  {
    QString msg, msg2;
    msg += i18n( "Operation complete." );
    msg += "\n\n";

    if( sc.numReplace == 0 )
    {
      msg2.sprintf( "%s", i18n( "No replacements were made." ) );
    }
    else if( sc.numReplace == 1 )
    {
      msg2.sprintf( "%s", i18n( "One replacement was made." ) );
    }
    else
    {
      msg2.sprintf( "%u %s", sc.numReplace, i18n("replacements were made.") );
    }
    msg += msg2;

    QMessageBox::information(
      topLevelWidget(),
      i18n("Find and Replace"),
      msg,
      i18n( "&OK" ),
      "",
      "",
      0, 0 );
  }
}


void CHexEditorWidget::insertPattern( void )
{
  if( mInsertDialog == 0 )
  {
    mInsertDialog = new CInsertDialog( 0, "insert" );
    if( mInsertDialog == 0 ) { return; }
    mInsertDialog->setCaption( i18n("Insert pattern") );
    connect( mInsertDialog, SIGNAL(execute( SInsertData & )), 
	     mHexView, SLOT(insert( SInsertData & )) );
  }
  if( mInsertDialog->isVisible() == false )
  {
    centerDialog( mInsertDialog, topLevelWidget() );
    mInsertDialog->defaultFocus();
    mInsertDialog->show();
  }
  else
  {
    mInsertDialog->raise();
  }
}


void CHexEditorWidget::encoding( void )
{
  QString msg( i18n("Not available yet!") );     
  msg += "\n";
  msg += i18n("Define your own encoding");
      
  QMessageBox::warning(
    topLevelWidget(),
    i18n("Encoding"),
    msg,
    i18n("&OK"),
    "",
    "",
    0, 0);
}

void CHexEditorWidget::strings( void )
{
  QString msg( i18n("Not available yet!") );     
  msg += "\n";
  msg += i18n("Locate strings (see \"man strings\") in document");
      
  QMessageBox::warning(
    topLevelWidget(),
    i18n("Extract strings"),
    msg,
    i18n("&OK"),
    "",
    "",
    0, 0);
}

void CHexEditorWidget::recordView( void )
{
  QString msg( i18n("Not available yet!") );     
  msg += "\n";
  msg += i18n("Make a record (structure) and fill it with data from document");
      
  QMessageBox::warning(
    topLevelWidget(),
    i18n("Record viewer"),
    msg,
    i18n("&OK"),
    "",
    "",
    0, 0);
}

void CHexEditorWidget::filter( void )
{
  QString msg( i18n("Not available yet!") );     
  msg += "\n";
  msg += i18n("Apply a binary filter on an area in document");
      
  QMessageBox::warning(
    topLevelWidget(),
    i18n("Binary filter"),
    msg,
    i18n("&OK"),
    "",
    "",
    0, 0);
}

void CHexEditorWidget::statistics( void )
{
  QString msg( i18n("Not available yet!") );     
  msg += "\n";
  msg += i18n("Document statistics (character frequency etc.)");
      
  QMessageBox::warning(
    topLevelWidget(),
    i18n("Statistics"),
    msg,
    i18n("&OK"),
    "",
    "",
    0, 0);
}


void CHexEditorWidget::options( void )
{
  if( mOptionDialog == 0 )
  {
    const char *title = i18n( "General Preferences" );
    mOptionDialog = new COptionDialog( title, 0 );
    if( mOptionDialog == 0 ) { return; }

    connect( mOptionDialog, SIGNAL(lineSizeChoise(const SDisplayLine &)),
	     SLOT(setLineSize(const SDisplayLine &)) );
    connect( mOptionDialog, SIGNAL(layoutChoise(const SDisplayLayout &)),
	     SLOT(setLayout(const SDisplayLayout &)) );
    connect( mOptionDialog, SIGNAL(fontChoise(const SDisplayFont &)),
	     SLOT(setFont(const SDisplayFont &)) );
    connect( mOptionDialog, SIGNAL(colorChoise(const SDisplayColor &)),
	     SLOT(setColor(const SDisplayColor &)) );
    connect( mOptionDialog, SIGNAL(cursorChoise(const SDisplayCursor &)),
	     SLOT(setCursor(const SDisplayCursor &)) );
    connect( mOptionDialog, SIGNAL(miscChoise(const SDisplayMisc &)),
	     SLOT(setMisc(const SDisplayMisc &)) );

    mOptionDialog->setState( mDisplayState );
    centerDialog( mOptionDialog, topLevelWidget() );
  }
  if( mOptionDialog->isVisible() == false )
  {
    mOptionDialog->setState( mDisplayState );
    centerDialog( mOptionDialog, topLevelWidget() );
    mOptionDialog->show();
  }
  else
  {
    mOptionDialog->raise();
  }
}



void CHexEditorWidget::favorites( void )
{
  QString msg( i18n("Not available yet!") );     
  msg += "\n";
  msg += i18n("Save or retrive your favorite layout");
      
  QMessageBox::warning(
    topLevelWidget(),
    i18n("Favorites"),
    msg,
    i18n("&OK"),
    "",
    "",
    0, 0);
}







int CHexEditorWidget::readURL( const QString &url, bool insert )
{
  //
  // 1) Make sure there is data in the url
  //
  if( url.isEmpty() == true )
  {
    return( Err_EmptyArgument );
  }

  //
  // 2) Verify that the url is valid URL string. If not, try to repair it.
  // This will work it the url contains a name of a file in the
  // current directory.
  //
  KURL kurl( url.data() );
  if( kurl.isMalformed() == true )
  {
    QString tmp( "file:" );
    if( *url.data() != '/' )
    {
      tmp += QDir::currentDirPath() + "/";
    }
    tmp += url;
    kurl.parse( tmp );

    if( kurl.isMalformed() == true )
    {
      QString msg( i18n("Malformed URL") );     
      msg += "\n" + kurl.url();
      
      QMessageBox::warning(
	topLevelWidget(),
        i18n("Sorry"),
        msg,
        i18n("&OK"),
        "",
        "",
        0, 0);
      return( Err_IllegalArgument );
    }
  }

  //
  // 3) Load the file. Since this is an url we check if this is a local
  // or a "remote" file. The local file can be opened as a regular file.
  //
  if( kurl.isLocalFile() == true )
  {
    QString path( kurl.path() );
    KURL::decodeURL( path );
    bool success = readFile( path, kurl.url(), insert );
    return( success == true ? Err_Success : Err_ReadFailed );
  }
  else
  {
    KFM *kfm = CreateKFMConnection();
    if( kfm == 0 )
    {
      return( Err_NoMemory );
    }

    mKfmURL    = url.data();
    mKfmTMP.sprintf( "/tmp/""khexedit%u", time(0) );
    mKfmAction = CHexEditorWidget::GET;

    connect( kfm, SIGNAL(finished()),
             this, SLOT(kfmFinished()) );
    connect( kfm, SIGNAL(error(int, const char *)),
             this, SLOT(kfmError(int, const char *)) );

    kfm->copy( url.data(), mKfmTMP );

  }
  return( Err_Success );

}


void CHexEditorWidget::writeURL( QString &url ) 
{
  KURL kurl( url );

  if( kurl.isLocalFile() == true )
  {
    QString name( kurl.path() );
    KURL::decodeURL( name );
    writeFile( name );
  } 
  else 
  {
    mKfmURL    = url.data();
    mKfmTMP.sprintf( "/tmp/""khexedit%u", time(0) );
    mKfmAction = CHexEditorWidget::PUT;

    if( writeFile( mKfmTMP ) == false ) 
    {
      return;
    }

    KFM *kfm = CreateKFMConnection();
    if( kfm == 0 )
    {
      // FIXME unlink here
      return;
    }

    connect( kfm, SIGNAL(finished()),
             this, SLOT(kfmFinished()) );
    connect( kfm, SIGNAL(error(int, const char *)),
             this, SLOT(kfmError(int, const char *)) );

    kfm->copy( mKfmTMP, url.data() );
  }
}



bool CHexEditorWidget::readFile( const QString &diskPath, const QString &url,
				 bool insert ) 
{

  QFileInfo info( diskPath.data() );
  if( info.exists() == FALSE )
  {
    QMessageBox::warning(
      topLevelWidget(),
      i18n("Read"),
      i18n("The specified file does not exist."),
      i18n("&OK"),
      "",
      "",
      0,0);
    return( false );
  }

  if( info.isDir() == TRUE ) 
  {
    QMessageBox::warning(
      topLevelWidget(),
      i18n("Read"),
      i18n("You have specified a directory."),
      i18n("&OK"),
      "",
      "",
      0,0);
    return( false );
  }

  if( info.isReadable() == FALSE ) 
  {
    QMessageBox::warning(
      topLevelWidget(),
      i18n("Read"),
      i18n("You do not have read permission to this file."),
      i18n("&OK"),
      "",
      "",
      0,0);
    return( false );
  }

  QFile file( diskPath.data() );
  if( file.open( IO_ReadOnly | IO_Raw ) == FALSE ) 
  {
    QMessageBox::warning(
      topLevelWidget(),
      i18n("Read"),
      i18n("An error occured while trying to open file."),
      i18n("&OK"),
      "",
      "",
      0,0);
    return( false );
  }

  int errCode = prepareProgressData( insert == true ? pg_insert: pg_read );
  if( errCode == Err_Success ) 
  {
    if( insert == true )
    {
      errCode = mHexView->insertFile( file, mProgressData );
    }
    else
    {
      errCode = mHexView->readFile( file, url, mProgressData );
    }
  }

  QString header = insert == true ? i18n("Insert") : i18n("Read");
  if( errCode != Err_Success )
  {
    QMessageBox::warning(
      topLevelWidget(),
      header.data(),
      hexError( errCode ),
      i18n("&OK"),
      "",
      "",
      0, 0 );
  }

  file.close();
  return( errCode == Err_Success || errCode == Err_Busy ? true : false );
}




bool CHexEditorWidget::writeFile( QString &diskPath ) 
{
  QFileInfo info( diskPath.data() );
  if( info.exists() == true )
  { 
    if( info.isDir() == true ) 
    {
      QMessageBox::warning(
        topLevelWidget(),
	i18n("Write failure"),
	i18n("You have specified a directory."),
	i18n("&OK"),
	"",
	"",
	0, 0 );
      return( false );
    }

    if( info.isWritable() == false ) 
    {
      QMessageBox::warning(
        topLevelWidget(),
	i18n("Write failure"),
	i18n("You do not have write permission."),
	i18n("&OK"),
	"",
	"",
	0, 0 );
      return( false );
    }
  }

  QFile file( diskPath.data() );
  if( file.open( IO_WriteOnly | IO_Raw ) == false ) 
  {
    QMessageBox::warning(
      topLevelWidget(),
      i18n("Write failure"),
      i18n("An error occured while trying to open file."),
      i18n("&OK"),
      "",
      "",
      0, 0 );
    return( false );
  }

  int errCode = prepareProgressData( pg_write );
  if( errCode == Err_Success )
  {
    errCode = mHexView->writeFile( file, mProgressData );
  }
  if( errCode != Err_Success )
  {
    QMessageBox::warning(
      topLevelWidget(),
      i18n("Write failure"),
      hexError( errCode ),
      i18n("&OK"),
      "",
      "",
      0,0 );
  }

  /*
  int errCode = mHexView->writeFile( file );
  if( errCode != Err_Success )
  {
    QMessageBox::warning(
      topLevelWidget(),
      i18n("Writing failed"),
      i18n("Could not write file."),
      i18n("&OK"),
      "",
      "",
      0,0 );
    return( false );
  }
  */

  file.close();
  return( true );
}






KFM *CHexEditorWidget::CreateKFMConnection( void )
{
  if( mKfm != 0 )
  {
    QString msg;
    msg += i18n( "An internet job is already in progress." );
    msg += "\n\n";
    msg += i18n( "Please wait until it has completed" );
    msg += "\n";
    msg += i18n( "or stop it." );

    QMessageBox::information(
      topLevelWidget(),
      i18n("Operation failed"),
      i18n( msg ),
      i18n("&OK"),
      "",
      "",
      0,0);
    return( 0 );
  }

  mKfm = new KFM;
  if( mKfm == 0 )
  {
    QString msg;
    msg += i18n( "Could not create KFM object." );
    msg += "\n";
    msg += i18n( "Insufficient memory." );

    QMessageBox::critical(
      topLevelWidget(),
      i18n("Operation failed"),
      msg.data(),
      i18n("&OK"),
      "",
      "",
      0,0);
    return( 0 );
  }

  if( mKfm->isOK() == false ) 
  {
    QString msg;
    msg += i18n( "Could not create KFM object." );
    msg += "\n";
    msg += i18n( "Object is not valid." );

    QMessageBox::warning(
      topLevelWidget(),
      i18n("Operation failed"),
      msg.data(),
      i18n("&OK"),
      "",
      "",
      0,0);
      delete mKfm; mKfm = 0;
      return( 0 );
  }

  
  return( mKfm ); 
}

void CHexEditorWidget::DestroyKFMConnection( void )
{
  delete mKfm; mKfm = 0;
}



void CHexEditorWidget::kfmFinished( void ) 
{
  if( mKfmAction == CHexEditorWidget::GET )
  {
    readFile( mKfmTMP, mKfmURL, false );
  }
  else if( mKfmAction == CHexEditorWidget::PUT )
  {
  }

  unlink( mKfmTMP.data() );
  DestroyKFMConnection();
}


void CHexEditorWidget::kfmError( int errCode, const char *errString ) 
{
  QString msg;
  msg += i18n( "KFM error" );
  msg += "\n";
  msg += errString;
  
  QMessageBox::critical(
    topLevelWidget(),
    i18n("Transfer failed"),
    msg.data(),
    i18n( "&OK" ),
    "",
    "",
    0,0 );
}

CHexBuffer *CHexEditorWidget::documentItem( const QString &url )
{ 
  for( CHexBuffer *hexBuffer = mDocumentList.first(); hexBuffer != 0; 
       hexBuffer = mDocumentList.next() )
  {
    if( hexBuffer->url() == url )
    {
      return( hexBuffer );
    }
  }
  
  return( 0 );
}


bool CHexEditorWidget::createBuffer( void )
{
  if( mHexView->valid() == false )
  {
    //
    // A buffer is already present, but it contains no data 
    //
    return( true );
  }

  CHexBuffer *hexBuffer = new CHexBuffer;
  if( hexBuffer == 0 ) 
  { 
    QString msg;
    msg += i18n( "Can not create text buffer." );
    msg += "\n";
    msg += i18n( "Insufficient memory." );

    QMessageBox::critical(
    topLevelWidget(),
    i18n("Loading failed" ),
    msg.data(),
    i18n( "&OK" ),
    "",
    "",
    0,0 );
    return( false ); 
  }
  
  mDocumentList.append( hexBuffer );
  mHexView->setBuffer( hexBuffer );

  return( true );
}


void CHexEditorWidget::removeBuffer( void )
{
  mHexView->closeFile();

  if( mDocumentList.count() > 1 )
  {
    CHexBuffer *prev = 0;
    CHexBuffer *curr = mDocumentList.first();
    for( ; curr != 0; curr = mDocumentList.next() )
    {
      if( curr == mHexView->hexBuffer() )
      {
	CHexBuffer *ptr = prev != 0 ? prev : mDocumentList.next();
	if( ptr != 0 )
	{
	  mHexView->setBuffer( ptr );
	  mDocumentList.remove( curr );
	  break;
	}
      }
      prev = curr;
    }
  }

}


bool CHexEditorWidget::modified( void )
{
  for( CHexBuffer *hexBuffer = mDocumentList.first(); hexBuffer != 0; 
       hexBuffer = mDocumentList.next() )
  {
    if( hexBuffer->modified() == true )
    {
      return( true );
    }
  }
  
  return( false );
}




int CHexEditorWidget::prepareProgressData( EProgressMode mode )
{
  if( mode > pg_encode )
  {
    return( Err_IllegalMode );
  }

  if( mProgressBusy == true )
  {
    return( Err_Busy );
  }


  mProgressMode = mode;
  mProgressBusy = true;
  mProgressStop = false;

  static const char *names[] = 
  {
    i18n("Reading"),
    i18n("Writing"),
    i18n("Inserting"),
    i18n("Printing"),
    i18n("Encoding"),
  };

  mProgressData.define( progressReceiver, this );
  emit setProgressText( QString(names[mode]) );
  emit operationChanged( true );

  return( Err_Success );
}


int CHexEditorWidget::progressReceiver( void *clientData, SProgressData &pd )
{
  if( clientData != 0 )
  {
    int errCode = ((CHexEditorWidget*)clientData)->progressParse( pd );
    return( errCode );
  }
  else
  {
    return( Err_Success );
  }
}


int CHexEditorWidget::progressParse( const SProgressData &pd )
{
  if( pd.valid() == 0 )
  {
    emit enableProgressText( false );
    emit setProgress( 0 );
    emit operationChanged( false );
    mProgressBusy = false;
    kapp->processEvents();
    return( Err_Success );
  }
  else if( pd.useFraction == 1 )
  {
    emit enableProgressText( true );
    emit setProgress( (int)(100.0 * pd.fraction ) );
    kapp->processEvents();
  }
  else
  {
    emit enableProgressText( true );
    emit setProgress( pd.curPage, pd.maxPage );
    kapp->processEvents();
  }

  if( mProgressStop == false )
  {
    return( Err_Success );
  }

  QString msg, header;
  switch( mProgressMode )
  {
    case pg_read:
      header = i18n("Read");
      msg = i18n("Do you really want to cancel reading?");
    break;

    case pg_write:
      msg  = i18n("Do you really want to cancel writing?");
      msg += "\n";
      msg += i18n("WARNING: Canceling can corrupt your data on disk");
    break;

    case pg_insert:
      header = i18n("Insert");
      msg = i18n("Do you really want to cancel inserting?");
    break;

    case pg_print:
      header = i18n("Print");
      msg = i18n("Do you really want to cancel printing?");
    break;

    case pg_encode:
      header = i18n("Encode");
      msg = i18n("Do you really want to cancel encoding?");
    break;

    default:
      return( Err_Success );
    break;
      
  }

  int reply = QMessageBox::information( 
    topLevelWidget(),
    header.data(),
    msg.data(),
    i18n( "&Yes" ),
    i18n( "&No" ),
    0, 
    0, 1 );
  mProgressStop = false;
  return( reply == 0 ? Err_Stop : Err_Success );

}



